from dataclasses import dataclass, field, InitVar
from datetime import datetime
from enum import Enum, unique
from typing import cast, List, Optional, Set, Any

from .ballot import _list_eq
from .election_object_base import ElectionObjectBase
from .group import Q, P, R, G, ElementModQ, ElementModP
from .hash import CryptoHashable, hash_elems
from .logs import log_warning
from .serializable import Serializable
from .utils import get_optional


@unique
class ElectionType(Enum):
    """
    enumerations for the `ElectionReport` entity
    see: https://developers.google.com/elections-data/reference/election-type
    """

    unknown = 0
    general = 1
    partisan_primary_closed = 2
    partisan_primary_open = 3
    primary = 4
    runoff = 5
    special = 6
    other = 7


@unique
class ReportingUnitType(Enum):
    """
    Enumeration for the type of geopolitical unit
    see: https://developers.google.com/elections-data/reference/reporting-unit-type
    """

    unknown = 0
    ballot_batch = 1
    ballot_style_area = 2
    borough = 3
    city = 4
    city_council = 5
    combined_precinct = 6
    congressional = 7
    country = 8
    county = 9
    county_council = 10
    drop_box = 11
    judicial = 12
    municipality = 13
    polling_place = 14
    precinct = 15
    school = 16
    special = 17
    split_precinct = 18
    state = 19
    state_house = 20
    state_senate = 21
    township = 22
    utility = 23
    village = 24
    vote_center = 25
    ward = 26
    water = 27
    other = 28


@unique
class VoteVariationType(Enum):
    """
    Enumeration for contest algorithm or rules in the `Contest` entity
    see: https://developers.google.com/elections-data/reference/vote-variation
    """

    unknown = 0
    one_of_m = 1
    approval = 2
    borda = 3
    cumulative = 4
    majority = 5
    n_of_m = 6
    plurality = 7
    proportional = 8
    range = 9
    rcv = 10
    super_majority = 11
    other = 12


@dataclass(eq=True, unsafe_hash=True)
class AnnotatedString(Serializable, CryptoHashable):
    """
    Use this as a type for character strings.
    See: https://developers.google.com/elections-data/reference/annotated-string
    """

    annotation: str = field(default="")
    value: str = field(default="")

    def crypto_hash(self) -> ElementModQ:
        """
        A hash representation of the object
        """
        return hash_elems(self.annotation, self.value)


@dataclass(eq=True, unsafe_hash=True)
class Language(Serializable, CryptoHashable):
    """
    The ISO-639 language
    see: https://en.wikipedia.org/wiki/ISO_639
    """

    value: str
    language: str = field(default="en")

    def crypto_hash(self) -> ElementModQ:
        """
        A hash representation of the object
        """
        return hash_elems(self.value, self.language)


@dataclass(eq=True, unsafe_hash=True)
class InternationalizedText(Serializable, CryptoHashable):
    """
    Data entity used to represent multi-national text. Use when text on a ballot contains multi-national text.
    See: https://developers.google.com/elections-data/reference/internationalized-text
    """

    text: List[Language] = field(default_factory=lambda: [])

    def crypto_hash(self) -> ElementModQ:
        """
        A hash representation of the object
        """
        return hash_elems(self.text)


@dataclass(eq=True, unsafe_hash=True)
class ContactInformation(Serializable, CryptoHashable):
    """
    For defining contact information about objects such as persons, boards of authorities, and organizations.
    See: https://developers.google.com/elections-data/reference/contact-information
    """

    address_line: Optional[List[str]] = field(default=None)
    email: Optional[List[AnnotatedString]] = field(default=None)
    phone: Optional[List[AnnotatedString]] = field(default=None)
    name: Optional[str] = field(default=None)

    def crypto_hash(self) -> ElementModQ:
        """
        A hash representation of the object
        """
        return hash_elems(self.name, self.address_line, self.email, self.phone)


@dataclass(eq=True, unsafe_hash=True)
class GeopoliticalUnit(ElectionObjectBase, CryptoHashable):
    """
    Use this entity for defining geopolitical units such as cities, districts, jurisdictions, or precincts, 
    for the purpose of associating contests, offices, vote counts, or other information with the geographies.
    See: https://developers.google.com/elections-data/reference/gp-unit
    """

    name: str
    type: ReportingUnitType
    contact_information: Optional[ContactInformation] = field(default=None)

    def crypto_hash(self) -> ElementModQ:
        """
        A hash representation of the object
        """
        return hash_elems(
            self.object_id, self.name, str(self.type), self.contact_information
        )


@dataclass(eq=True, unsafe_hash=True)
class BallotStyle(ElectionObjectBase, CryptoHashable):
    """
    A BallotStyle works as a key to uniquely specify a set of contests. See also `ContestDescription`.
    """

    geopolitical_unit_ids: Optional[List[str]] = field(default=None)
    party_ids: Optional[List[str]] = field(default=None)
    image_uri: Optional[str] = field(default=None)

    def crypto_hash(self) -> ElementModQ:
        """
        A hash representation of the object
        """
        return hash_elems(
            self.object_id, self.geopolitical_unit_ids, self.party_ids, self.image_uri
        )


@dataclass(eq=True, unsafe_hash=True)
class Party(ElectionObjectBase, CryptoHashable):
    """
    Use this entity to describe a political party that can then be referenced from other entities.
    See: https://developers.google.com/elections-data/reference/party
    """

    ballot_name: InternationalizedText = field(default=InternationalizedText())
    abbreviation: Optional[str] = field(default=None)
    color: Optional[str] = field(default=None)
    logo_uri: Optional[str] = field(default=None)

    def get_party_id(self) -> str:
        """
        Returns the object identifier associated with the Party.
        """
        return self.object_id

    def crypto_hash(self) -> ElementModQ:
        """
        A hash representation of the object
        """
        return hash_elems(
            self.object_id,
            self.ballot_name,
            self.abbreviation,
            self.color,
            self.logo_uri,
        )


@dataclass(eq=True, unsafe_hash=True)
class Candidate(ElectionObjectBase, CryptoHashable):
    """
    Entity describing information about a candidate in a contest. 
    See: https://developers.google.com/elections-data/reference/candidate
    Note: The ElectionGuard Data Spec deviates from the NIST model in that 
    selections for any contest type are considered a "candidate".
    for instance, on a yes-no referendum contest, two `candidate` objects
    would be included in the model to represent the `affirmative` and `negative`
    selections for the contest.  See the wiki, readme's, and tests in this repo for more info
    """

    ballot_name: InternationalizedText = field(default=InternationalizedText())
    party_id: Optional[str] = field(default=None)
    image_uri: Optional[str] = field(default=None)
    is_write_in: Optional[bool] = field(default=None)

    def get_candidate_id(self) -> str:
        """
        Given a `Candidate`, returns a "candidate ID", which is used in other ElectionGuard structures.
        """
        return self.object_id

    def crypto_hash(self) -> ElementModQ:
        """
        A hash representation of the object
        """
        return hash_elems(
            self.object_id, self.ballot_name, self.party_id, self.image_uri
        )


@dataclass(eq=True, unsafe_hash=True)
class SelectionDescription(ElectionObjectBase, CryptoHashable):
    """
    Data entity for the ballot selections in a contest, 
    for example linking candidates and parties to their vote counts.
    See: https://developers.google.com/elections-data/reference/ballot-selection
    Note: The ElectionGuard Data Spec deviates from the NIST model in that
    there is no difference for different types of selections. 
    The ElectionGuard Data Spec deviates from the NIST model in that
    `sequence_order` is a required field since it is used for ordering selections 
    in a contest to ensure various encryption primitives are deterministic.
    For a given election, the sequence of selections displayed to a user may be different
    however that information is not captured by default when encrypting a specific ballot.
    """

    candidate_id: str
    sequence_order: int
    """
    Used for ordering selections in a contest to ensure various encryption primitives are deterministic.
    The sequence order must be unique and should be representative of how the contests are represnted
    on a "master" ballot in an external system.  The sequence order is not required to be in the order 
    in which they are displayed to a voter.  Any acceptable range of integer values may be provided.
    """

    def crypto_hash(self) -> ElementModQ:
        """
        A hash representation of the object
        """
        return hash_elems(self.object_id, self.sequence_order, self.candidate_id)


@dataclass(unsafe_hash=True)
class ContestDescription(ElectionObjectBase, CryptoHashable):
    """
    Use this data entity for describing a contest and linking the contest 
    to the associated candidates and parties.
    See: https://developers.google.com/elections-data/reference/contest
    Note: The ElectionGuard Data Spec deviates from the NIST model in that
    `sequence_order` is a required field since it is used for ordering selections 
    in a contest to ensure various encryption primitives are deterministic.
    For a given election, the sequence of contests displayed to a user may be different
    however that information is not captured by default when encrypting a specific ballot.
    """

    electoral_district_id: str
    sequence_order: int
    """
    Used for ordering contests in a ballot to ensure various encryption primitives are deterministic.
    The sequence order must be unique and should be representative of how the contests are represnted
    on a "master" ballot in an external system.  The sequence order is not required to be in the order 
    in which they are displayed to a voter.  Any acceptable range of integer values may be provided.
    """

    vote_variation: VoteVariationType

    # Number of candidates that are elected in the contest ("n" of n-of-m).
    # Note: a referendum is considered a specific case of 1-of-m in ElectionGuard
    number_elected: int

    # Maximum number of votes/write-ins per voter in this contest. Used in cumulative voting
    # to indicate how many total votes a voter can spread around. In n-of-m elections, this will
    # be None.
    votes_allowed: Optional[int]

    # Name of the contest, not necessarily as it appears on the ballot.
    name: str

    # For associating a ballot selection for the contest, i.e., a candidate, a ballot measure.
    ballot_selections: List[SelectionDescription] = field(default_factory=lambda: [])

    # Title of the contest as it appears on the ballot.
    ballot_title: Optional[InternationalizedText] = field(default=None)

    # Subtitle of the contest as it appears on the ballot.
    ballot_subtitle: Optional[InternationalizedText] = field(default=None)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, ContestDescription)
            and self.electoral_district_id == other.electoral_district_id
            and self.sequence_order == other.sequence_order
            and self.votes_allowed == other.votes_allowed
            and self.number_elected == other.number_elected
            and self.votes_allowed == other.votes_allowed
            and self.name == other.name
            and _list_eq(self.ballot_selections, other.ballot_selections)
            and self.ballot_title == other.ballot_title
            and self.ballot_subtitle == other.ballot_subtitle
        )

    def crypto_hash(self) -> ElementModQ:
        """
        Given a ContestDescription, deterministically derives a "hash" of that contest,
        suitable for use in ElectionGuard's "base hash" values, and for validating that
        either a plaintext or encrypted voted context and its corresponding contest
        description match up.
        """
        # remove any placeholders from the hash mechanism
        return hash_elems(
            self.object_id,
            self.sequence_order,
            self.electoral_district_id,
            str(self.vote_variation),
            self.ballot_title,
            self.ballot_subtitle,
            self.name,
            self.number_elected,
            self.votes_allowed,
            self.ballot_selections,
        )

    def is_valid(self) -> bool:
        """
        Check the validity of the contest object by verifying its data
        """
        contest_has_valid_number_elected = self.number_elected <= len(
            self.ballot_selections
        )
        contest_has_valid_votes_allowed = (
            self.votes_allowed is None or self.number_elected <= self.votes_allowed
        )

        # verify the candidate_ids, selection object_ids, and sequence_ids are unique
        candidate_ids: Set[str] = set()
        selection_ids: Set[str] = set()
        sequence_ids: Set[int] = set()
        selection_count = 0
        expected_selection_count = len(self.ballot_selections)

        for selection in self.ballot_selections:
            selection_count += 1
            # validate the object_id
            if selection.object_id not in selection_ids:
                selection_ids.add(selection.object_id)
            # validate the sequence_order
            if selection.sequence_order not in sequence_ids:
                sequence_ids.add(selection.sequence_order)
            # validate the candidate id
            if selection.candidate_id not in candidate_ids:
                candidate_ids.add(selection.candidate_id)

        selections_have_valid_candidate_ids = (
            len(candidate_ids) == expected_selection_count
        )
        selections_have_valid_selection_ids = (
            len(selection_ids) == expected_selection_count
        )
        selections_have_valid_sequence_ids = (
            len(sequence_ids) == expected_selection_count
        )

        success = (
            contest_has_valid_number_elected
            and contest_has_valid_votes_allowed
            and selections_have_valid_candidate_ids
            and selections_have_valid_selection_ids
            and selections_have_valid_sequence_ids
        )

        if not success:
            log_warning(
                "Contest %s failed validation check: %s",
                self.object_id,
                str(
                    {
                        "contest_has_valid_number_elected": contest_has_valid_number_elected,
                        "contest_has_valid_votes_allowed": contest_has_valid_votes_allowed,
                        "selections_have_valid_candidate_ids": selections_have_valid_candidate_ids,
                        "selections_have_valid_selection_ids": selections_have_valid_selection_ids,
                        "selections_have_valid_sequence_ids": selections_have_valid_sequence_ids,
                    }
                ),
            )

        return success


@dataclass(eq=True, unsafe_hash=True)
class CandidateContestDescription(ContestDescription):
    """
    Use this entity to describe a contest that involves selecting one or more candidates.
    See: https://developers.google.com/elections-data/reference/contest
    Note: The ElectionGuard Data Spec deviates from the NIST model in that
    this subclass is used purely for convenience 
    """

    primary_party_ids: List[str] = field(default_factory=lambda: [])


@dataclass(eq=True, unsafe_hash=True)
class ReferendumContestDescription(ContestDescription):
    """
    Use this entity to describe a contest that involves selecting exactly one 'candidate'.
    See: https://developers.google.com/elections-data/reference/contest
    Note: The ElectionGuard Data Spec deviates from the NIST model in that
    this subclass is used purely for convenience
    """

    pass


@dataclass(eq=True, unsafe_hash=True)
class ContestDescriptionWithPlaceholders(ContestDescription):
    """
    ContestDescriptionWithPlaceholders is a `ContestDescription` with ElectionGuard `placeholder_selections`.
    (The ElectionGuard spec requires for n-of-m elections that there be *exactly* n counters that are one
    with the rest zero, so if a voter deliberately undervotes, one or more of the placeholder counters will
    become one. This allows the `ConstantChaumPedersenProof` to verify correctly for undervoted contests.)
    """

    placeholder_selections: List[SelectionDescription] = field(
        default_factory=lambda: []
    )

    def is_valid(self) -> bool:
        """
        Checks is contest description is valid
        :return: true if valid
        """
        contest_description_validates = super().is_valid()
        return (
            contest_description_validates
            and len(self.placeholder_selections) == self.number_elected
        )

    def is_placeholder(self, selection: SelectionDescription) -> bool:
        """
        Checks is contest description is placeholder
        :return: true if placeholder
        """
        return selection in self.placeholder_selections

    def selection_for(self, selection_id: str) -> Optional[SelectionDescription]:
        """
        Gets the description for a particular id
        :param selection_id: Id of Selection
        :return: description
        """
        matching_selections = list(
            filter(lambda i: i.object_id == selection_id, self.ballot_selections)
        )

        if any(matching_selections):
            return matching_selections[0]

        matching_placeholders = list(
            filter(lambda i: i.object_id == selection_id, self.placeholder_selections)
        )

        if any(matching_placeholders):
            return matching_placeholders[0]
        else:
            return None


@dataclass(unsafe_hash=True)
class ElectionDescription(Serializable, CryptoHashable):
    """
    Use this entity for defining the structure of the election and associated 
    information such as candidates, contests, and vote counts.  This class is
    based on the NIST Election Common Standard Data Specification.  Some deviations
    from the standard exist.

    This structure is considered an immutable input object and should not be changed
    through the course of an election, as it's hash representation is the basis for all
    other hash representations within an ElectionGuard election context.

    See: https://developers.google.com/elections-data/reference/election
    """

    election_scope_id: str
    type: ElectionType
    start_date: datetime
    end_date: datetime
    geopolitical_units: List[GeopoliticalUnit]
    parties: List[Party]
    candidates: List[Candidate]
    contests: List[ContestDescription]
    ballot_styles: List[BallotStyle]
    name: Optional[InternationalizedText] = field(default=None)
    contact_information: Optional[ContactInformation] = field(default=None)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, ElectionDescription)
            and self.election_scope_id == other.election_scope_id
            and self.type == other.type
            and self.start_date == other.start_date
            and self.end_date == other.end_date
            and _list_eq(self.geopolitical_units, other.geopolitical_units)
            and _list_eq(self.parties, other.parties)
            and _list_eq(self.candidates, other.candidates)
            and _list_eq(self.contests, other.contests)
            and _list_eq(self.ballot_styles, other.ballot_styles)
            and self.name == other.name
            and self.contact_information == other.contact_information
        )

    def crypto_hash(self) -> ElementModQ:
        """
        Returns a hash of the metadata components of the election
        """

        return hash_elems(
            self.election_scope_id,
            str(self.type),
            self.start_date.isoformat(),
            self.end_date.isoformat(),
            self.name,
            self.contact_information,
            self.geopolitical_units,
            self.parties,
            self.parties,
            self.contests,
            self.ballot_styles,
        )

    def is_valid(self) -> bool:
        """
        Verifies the dataset to ensure it is well-formed
        """
        gp_unit_ids: Set[str] = set()
        ballot_style_ids: Set[str] = set()
        party_ids: Set[str] = set()
        candidate_ids: Set[str] = set()
        contest_ids: Set[str] = set()
        selection_ids: Set[str] = set()

        # Validate GP Units
        for gp_unit in self.geopolitical_units:
            if gp_unit.object_id not in gp_unit_ids:
                gp_unit_ids.add(gp_unit.object_id)

        # fail if there are duplicates
        geopolitical_units_valid = len(gp_unit_ids) == len(self.geopolitical_units)

        # Validate Ballot Styles
        ballot_styles_have_valid_gp_unit_ids = True

        for style in self.ballot_styles:
            if style.object_id not in ballot_style_ids:
                ballot_style_ids.add(style.object_id)
            if style.geopolitical_unit_ids is None:
                ballot_styles_have_valid_gp_unit_ids = False
                break
            # validate associated gp unit ids
            for gp_unit_id in style.geopolitical_unit_ids:
                ballot_styles_have_valid_gp_unit_ids = (
                    ballot_styles_have_valid_gp_unit_ids and gp_unit_id in gp_unit_ids
                )

        ballot_styles_valid = (
            len(ballot_style_ids) == len(self.ballot_styles)
            and ballot_styles_have_valid_gp_unit_ids
        )

        # Validate Parties
        for party in self.parties:
            if party.object_id not in party_ids:
                party_ids.add(party.object_id)

        parties_valid = len(party_ids) == len(self.parties)

        # Validate Candidates
        candidates_have_valid_party_ids = True

        for candidate in self.candidates:
            if candidate.object_id not in candidate_ids:
                candidate_ids.add(candidate.object_id)
            # validate the associated party id
            candidates_have_valid_party_ids = candidates_have_valid_party_ids and (
                candidate.party_id is None or candidate.party_id in party_ids
            )

        candidates_have_valid_length = len(candidate_ids) == len(self.candidates)
        candidates_valid = (
            candidates_have_valid_length and candidates_have_valid_party_ids
        )

        # Validate Contests
        contests_validate_their_properties = True
        contests_have_valid_electoral_district_id = True
        candidate_contests_have_valid_party_ids = True

        contest_sequence_ids: Set[int] = set()

        for contest in self.contests:

            contests_validate_their_properties = (
                contests_validate_their_properties and contest.is_valid()
            )

            if contest.object_id not in contest_ids:
                contest_ids.add(contest.object_id)

            # validate the sequence order
            if contest.sequence_order not in contest_sequence_ids:
                contest_sequence_ids.add(contest.sequence_order)

            # validate the associated gp unit id
            contests_have_valid_electoral_district_id = (
                contests_have_valid_electoral_district_id
                and contest.electoral_district_id in gp_unit_ids
            )

            if isinstance(contest, CandidateContestDescription):
                candidate_contest = cast(CandidateContestDescription, contest)
                if candidate_contest.primary_party_ids is not None:
                    for primary_party_id in candidate_contest.primary_party_ids:
                        # validate the party ids
                        candidate_contests_have_valid_party_ids = (
                            candidate_contests_have_valid_party_ids
                            and primary_party_id in party_ids
                        )

        # TODO: ISSUE #55: verify that the contest sequence order set is in the proper order

        contests_have_valid_object_ids = len(contest_ids) == len(self.contests)
        contests_have_valid_sequence_ids = len(contest_sequence_ids) == len(
            self.contests
        )
        contests_valid = (
            contests_have_valid_object_ids
            and contests_have_valid_sequence_ids
            and contests_validate_their_properties
            and contests_have_valid_electoral_district_id
            and candidate_contests_have_valid_party_ids
        )

        success = (
            geopolitical_units_valid
            and ballot_styles_valid
            and parties_valid
            and candidates_valid
            and contests_valid
        )

        if not success:
            log_warning(
                "Election failed validation check: is_valid: %s",
                str(
                    {
                        "geopolitical_units_valid": geopolitical_units_valid,
                        "ballot_styles_valid": ballot_styles_valid,
                        "ballot_styles_have_valid_gp_unit_ids": ballot_styles_have_valid_gp_unit_ids,
                        "parties_valid": parties_valid,
                        "candidates_valid": candidates_valid,
                        "candidates_have_valid_length": candidates_have_valid_length,
                        "candidates_have_valid_party_ids": candidates_have_valid_party_ids,
                        "contests_valid": contests_valid,
                        "contests_have_valid_object_ids": contests_have_valid_object_ids,
                        "contests_have_valid_sequence_ids": contests_have_valid_sequence_ids,
                        "contests_validate_their_properties": contests_validate_their_properties,
                        "contests_have_valid_electoral_district_id": contests_have_valid_electoral_district_id,
                        "candidate_contests_have_valid_party_ids": candidate_contests_have_valid_party_ids,
                    }
                ),
            )
        return success


@dataclass(eq=True, unsafe_hash=True)
class InternalElectionDescription(object):
    """
    `InternalElectionDescription` is a subset of the `ElectionDescription` structure that specifies
    the components that ElectionGuard uses for conducting an election.  The key component is the
    `contests` collection, which applies placeholder selections to the `ElectionDescription` contests
    """

    description: InitVar[ElectionDescription] = None

    geopolitical_units: List[GeopoliticalUnit] = field(init=False)

    contests: List[ContestDescriptionWithPlaceholders] = field(init=False)

    ballot_styles: List[BallotStyle] = field(init=False)

    description_hash: ElementModQ = field(init=False)

    def __post_init__(self, description: ElectionDescription) -> None:
        object.__setattr__(self, "description_hash", description.crypto_hash())
        object.__setattr__(self, "geopolitical_units", description.geopolitical_units)
        object.__setattr__(self, "ballot_styles", description.ballot_styles)
        object.__setattr__(
            self, "contests", self._generate_contests_with_placeholders(description)
        )

    def contest_for(
        self, contest_id: str
    ) -> Optional[ContestDescriptionWithPlaceholders]:
        """
        Get contest by id
        :param contest_id: Contest id
        :return: Contest description or none
        """
        matching_contests = list(
            filter(lambda i: i.object_id == contest_id, self.contests)
        )

        if any(matching_contests):
            return matching_contests[0]
        else:
            return None

    def get_ballot_style(self, ballot_style_id: str) -> BallotStyle:
        """
        Get a ballot style for a specified ballot_style_id
        """
        style = list(
            filter(lambda i: i.object_id == ballot_style_id, self.ballot_styles)
        )[0]
        return style

    def get_contests_for(
        self, ballot_style_id: str
    ) -> List[ContestDescriptionWithPlaceholders]:
        """
        Get contests for a ballot style
        :param ballot_style_id: ballot style id
        :return: contest descriptions
        """
        style = self.get_ballot_style(ballot_style_id)
        if style.geopolitical_unit_ids is None:
            return list()
        gp_unit_ids = [gp_unit_id for gp_unit_id in style.geopolitical_unit_ids]
        contests = list(
            filter(lambda i: i.electoral_district_id in gp_unit_ids, self.contests)
        )
        return contests

    def _generate_contests_with_placeholders(
        self, description: ElectionDescription
    ) -> List[ContestDescriptionWithPlaceholders]:
        """
        For each contest, append the `number_elected` number 
        of placeholder selections to the end of the contest collection
        """
        contests: List[ContestDescriptionWithPlaceholders] = list()
        for contest in description.contests:
            placeholder_selections = generate_placeholder_selections_from(
                contest, contest.number_elected
            )
            contests.append(
                contest_description_with_placeholders_from(
                    contest, placeholder_selections
                )
            )

        return contests


@dataclass(eq=True, unsafe_hash=True)
class ElectionConstants(Serializable):
    """
    The constants for mathematical functions during the election. 
    """

    large_prime = P
    """large prime or p"""

    small_prime = Q
    """small prime or q"""

    cofactor = R
    """cofactor or r"""

    generator = G
    """generator or g"""


@dataclass(eq=True, unsafe_hash=True)
class CiphertextElectionContext(Serializable):
    """
    `CiphertextElectionContext` is the ElectionGuard representation of a specific election
    Note: The ElectionGuard Data Spec deviates from the NIST model in that
    this object includes fields that are populated in the course of encrypting an election
    Specifically, `crypto_base_hash`, `crypto_extended_base_hash` and `elgamal_public_key`
    are populated with election-specific information necessary for encrypting the election.
    Refer to the [Electionguard Specification](https://github.com/microsoft/electionguard) for more information.

    To make an instance of this class, don't construct it directly. Use
    `make_ciphertext_election_context` instead.
    """

    number_of_guardians: int
    """
    The number of guardians necessary to generate the public key
    """
    quorum: int
    """
    The quorum of guardians necessary to decrypt an election.  Must be less than `number_of_guardians`
    """

    # the `joint public key (K)` in the [ElectionGuard Spec](https://github.com/microsoft/electionguard/wiki)
    elgamal_public_key: ElementModP

    # The hash of the election metadata
    description_hash: ElementModQ

    # the `base hash code (𝑄)` in the [ElectionGuard Spec](https://github.com/microsoft/electionguard/wiki)
    crypto_base_hash: ElementModQ

    # the `extended base hash code (𝑄')` in the [ElectionGuard Spec](https://github.com/microsoft/electionguard/wiki)
    crypto_extended_base_hash: ElementModQ


def make_ciphertext_election_context(
    number_of_guardians: int,
    quorum: int,
    elgamal_public_key: ElementModP,
    description_hash: ElementModQ,
) -> CiphertextElectionContext:
    """
    Makes a CiphertextElectionContext object.

    :param number_of_guardians: The number of guardians necessary to generate the public key
    :param quorum: The quorum of guardians necessary to decrypt an election.  Must be less than `number_of_guardians`
    :param elgamal_public_key: the public key of the election
    :param description_hash: the hash of the election metadata
    """

    # What's a crypto_base_hash?
    # The metadata of this object are hashed together with the
    # - prime modulus (𝑝),
    # - subgroup order (𝑞),
    # - generator (𝑔),
    # - number of guardians (𝑛),
    # - decryption threshold value (𝑘),
    # to form a base hash code (𝑄) which will be incorporated
    # into every subsequent hash computation in the election.

    # What's a crypto_extended_base_hash?
    # Once the baseline parameters have been produced and confirmed,
    # all of the public guardian commitments 𝐾𝑖,𝑗 are hashed together
    # with the base hash 𝑄 to form an extended base hash 𝑄' that will
    # form the basis of subsequent hash computations.

    crypto_base_hash = hash_elems(
        P, Q, G, number_of_guardians, quorum, description_hash
    )
    crypto_extended_base_hash = hash_elems(crypto_base_hash, elgamal_public_key)
    return CiphertextElectionContext(
        number_of_guardians=number_of_guardians,
        quorum=quorum,
        elgamal_public_key=elgamal_public_key,
        description_hash=description_hash,
        crypto_base_hash=crypto_base_hash,
        crypto_extended_base_hash=crypto_extended_base_hash,
    )


def contest_description_with_placeholders_from(
    description: ContestDescription, placeholders: List[SelectionDescription]
) -> ContestDescriptionWithPlaceholders:
    """
    Generates a placeholder selection description
    :param description: contest description
    :param placeholders: list of placeholder descriptions of selections
    :return: a SelectionDescription or None
    """
    return ContestDescriptionWithPlaceholders(
        object_id=description.object_id,
        electoral_district_id=description.electoral_district_id,
        sequence_order=description.sequence_order,
        vote_variation=description.vote_variation,
        number_elected=description.number_elected,
        votes_allowed=description.votes_allowed,
        name=description.name,
        ballot_selections=description.ballot_selections,
        ballot_title=description.ballot_title,
        ballot_subtitle=description.ballot_subtitle,
        placeholder_selections=placeholders,
    )


def generate_placeholder_selection_from(
    contest: ContestDescription, use_sequence_id: Optional[int] = None
) -> Optional[SelectionDescription]:
    """
    Generates a placeholder selection description that is unique so it can be hashed

    :param use_sequence_id: an optional integer unique to the contest identifying this selection's place in the contest
    :return: a SelectionDescription or None
    """
    sequence_ids = [selection.sequence_order for selection in contest.ballot_selections]
    if use_sequence_id is None:
        # if no sequence order is specified, take the max
        use_sequence_id = max(*sequence_ids) + 1
    elif use_sequence_id in sequence_ids:
        log_warning(
            f"mismatched placeholder selection {use_sequence_id} already exists"
        )
        return None

    placeholder_object_id = f"{contest.object_id}-{use_sequence_id}"
    return SelectionDescription(
        f"{placeholder_object_id}-placeholder",
        f"{placeholder_object_id}-candidate",
        use_sequence_id,
    )


def generate_placeholder_selections_from(
    contest: ContestDescription, count: int
) -> List[SelectionDescription]:
    """
    Generates the specified number of placeholder selections in ascending sequence order from the max selection sequence orderf

    :param contest: ContestDescription for input
    :param count: optionally specify a number of placeholders to generate
    :return: a collection of `SelectionDescription` objects, which may be empty
    """
    max_sequence_order = max(
        [selection.sequence_order for selection in contest.ballot_selections]
    )
    selections: List[SelectionDescription] = list()
    for i in range(count):
        sequence_order = max_sequence_order + 1 + i
        selections.append(
            get_optional(generate_placeholder_selection_from(contest, sequence_order))
        )
    return selections



from typing import Callable, Optional, Tuple

from .chaum_pedersen import ChaumPedersenProof, make_chaum_pedersen
from .data_store import DataStore, ReadOnlyDataStore
from .election_object_base import ElectionObjectBase
from .elgamal import ElGamalCiphertext
from .group import (
    ElementModP,
    ElementModQ,
    int_to_q,
    mult_p,
    pow_q,
    pow_p,
    ONE_MOD_P,
    rand_q,
)
from .key_ceremony import (
    AuxiliaryKeyPair,
    AuxiliaryPublicKey,
    AuxiliaryDecrypt,
    AuxiliaryEncrypt,
    combine_election_public_keys,
    CeremonyDetails,
    CoefficientValidationSet,
    ElectionJointKey,
    ElectionKeyPair,
    ElectionPartialKeyBackup,
    ElectionPartialKeyChallenge,
    ElectionPartialKeyVerification,
    ElectionPublicKey,
    get_coefficient_validation_set,
    generate_election_key_pair,
    generate_election_partial_key_backup,
    generate_election_partial_key_challenge,
    generate_rsa_auxiliary_key_pair,
    PublicKeySet,
    verify_election_partial_key_backup,
    verify_election_partial_key_challenge,
)
from .logs import log_warning
from .rsa import rsa_encrypt, rsa_decrypt
from .types import GUARDIAN_ID
from .utils import get_optional


class Guardian(ElectionObjectBase):
    """
    Guardian of election responsible for safeguarding information and decrypting results
    """

    sequence_order: int
    ceremony_details: CeremonyDetails
    _auxiliary_keys: AuxiliaryKeyPair
    _election_keys: ElectionKeyPair
    _backups_to_share: DataStore[GUARDIAN_ID, ElectionPartialKeyBackup]
    """
    The collection of this guardian's partial key backups that will be shared to other guardians
    """

    # From Other Guardians
    _guardian_auxiliary_public_keys: DataStore[GUARDIAN_ID, AuxiliaryPublicKey]
    """
    The collection of other guardians' auxiliary public keys that are shared with this guardian
    """

    _guardian_election_public_keys: DataStore[GUARDIAN_ID, ElectionPublicKey]
    """
    The collection of other guardians' election public keys that are shared with this guardian
    """

    _guardian_election_partial_key_backups: DataStore[
        GUARDIAN_ID, ElectionPartialKeyBackup
    ]
    """
    The collection of other guardians' partial key backups that are shared with this guardian
    """

    _guardian_election_partial_key_verifications: DataStore[
        GUARDIAN_ID, ElectionPartialKeyVerification
    ]
    """
    The collection of other guardians' verifications that they shared their backups correctly
    """

    def __init__(
        self,
        id: str,
        sequence_order: int,
        number_of_guardians: int,
        quorum: int,
        nonce_seed: Optional[ElementModQ] = None,
    ) -> None:
        """
        Initialize a guardian with the specified arguments

        :param id: the unique identifier for the guardian
        :param sequence_order: a unique number in [0, 256) that identifies this guardian
        :param number_of_guardians: the total number of guardians that will participate in the election
        :param quorum: the count of guardians necessary to decrypt
        :param nonce_seed: an optional `ElementModQ` value that can be used to generate the `ElectionKeyPair`.  
                           It is recommended to only use this field for testing.
        """

        super().__init__(id)
        self.sequence_order = sequence_order
        self.set_ceremony_details(number_of_guardians, quorum)
        self._backups_to_share = DataStore[GUARDIAN_ID, ElectionPartialKeyBackup]()
        self._guardian_auxiliary_public_keys = DataStore[
            GUARDIAN_ID, AuxiliaryPublicKey
        ]()
        self._guardian_election_public_keys = DataStore[
            GUARDIAN_ID, ElectionPublicKey
        ]()
        self._guardian_election_partial_key_backups = DataStore[
            GUARDIAN_ID, ElectionPartialKeyBackup
        ]()
        self._guardian_election_partial_key_verifications = DataStore[
            GUARDIAN_ID, ElectionPartialKeyVerification
        ]()

        self.generate_auxiliary_key_pair()
        self.generate_election_key_pair(nonce_seed if nonce_seed is not None else None)

    def reset(self, number_of_guardians: int, quorum: int) -> None:
        """
        Reset guardian to initial state
        :param number_of_guardians: Number of guardians in election
        :param quorum: Quorum of guardians required to decrypt
        """
        self._backups_to_share.clear()
        self._guardian_auxiliary_public_keys.clear()
        self._guardian_election_public_keys.clear()
        self._guardian_election_partial_key_backups.clear()
        self._guardian_election_partial_key_verifications.clear()
        self.set_ceremony_details(number_of_guardians, quorum)
        self.generate_auxiliary_key_pair()
        self.generate_election_key_pair()

    def set_ceremony_details(self, number_of_guardians: int, quorum: int) -> None:
        """
        Set ceremony details for election
        :param number_of_guardians: Number of guardians in election
        :param quorum: Quorum of guardians required to decrypt
        """
        self.ceremony_details = CeremonyDetails(number_of_guardians, quorum)

    # Public Keys
    def share_public_keys(self) -> PublicKeySet:
        """
        Share public election and auxiliary keys for guardian
        :return: Public set of election and auxiliary keys
        """
        return PublicKeySet(
            self.object_id,
            self.sequence_order,
            self._auxiliary_keys.public_key,
            self._election_keys.key_pair.public_key,
            self._election_keys.proof,
        )

    def save_guardian_public_keys(self, public_key_set: PublicKeySet) -> None:
        """
        Save public election and auxiliary keys for another guardian
        :param public_key_set: Public set of election and auxiliary keys
        """
        self.save_auxiliary_public_key(
            AuxiliaryPublicKey(
                public_key_set.owner_id,
                public_key_set.sequence_order,
                public_key_set.auxiliary_public_key,
            )
        )
        self.save_election_public_key(
            ElectionPublicKey(
                public_key_set.owner_id,
                public_key_set.election_public_key_proof,
                public_key_set.election_public_key,
            ),
        )

    def all_public_keys_received(self) -> bool:
        """
        True if all election and auxiliary public keys have been received.
        :return: All election and auxiliary public keys backups received
        """
        return (
            self.all_auxiliary_public_keys_received()
            and self.all_election_public_keys_received()
        )

    def generate_auxiliary_key_pair(
        self,
        generate_auxiliary_key_pair: Callable[
            [], AuxiliaryKeyPair
        ] = generate_rsa_auxiliary_key_pair,
    ) -> None:
        """
        Generate auxiliary key pair
        :param generate_auxiliary_key_pair: Function to generate auxiliary key pair
        """
        self._auxiliary_keys = generate_auxiliary_key_pair()
        self.save_auxiliary_public_key(self.share_auxiliary_public_key())

    def share_auxiliary_public_key(self) -> AuxiliaryPublicKey:
        """
        Share auxiliary public key with another guardian
        :return: Auxiliary Public Key
        """
        return AuxiliaryPublicKey(
            self.object_id, self.sequence_order, self._auxiliary_keys.public_key
        )

    def save_auxiliary_public_key(self, key: AuxiliaryPublicKey) -> None:
        """
        Save a guardians auxiliary public key
        :param key: Auxiliary public key
        """
        self._guardian_auxiliary_public_keys.set(key.owner_id, key)

    def all_auxiliary_public_keys_received(self) -> bool:
        """
        True if all auxiliary public keys have been received.
        :return: All auxiliary public keys backups received
        """
        return (
            len(self._guardian_auxiliary_public_keys)
            == self.ceremony_details.number_of_guardians
        )

    def guardian_auxiliary_public_keys(
        self,
    ) -> ReadOnlyDataStore[GUARDIAN_ID, AuxiliaryPublicKey]:
        """
        Get a read-only view of the auxiliary public keys provided to this Guardian
        """
        return ReadOnlyDataStore(self._guardian_auxiliary_public_keys)

    def generate_election_key_pair(self, nonce: ElementModQ = None) -> None:
        """
        Generate election key pair for encrypting/decrypting election
        """
        self._election_keys = generate_election_key_pair(
            self.ceremony_details.quorum, nonce
        )
        self.save_election_public_key(self.share_election_public_key())

    def share_election_public_key(self) -> ElectionPublicKey:
        """
        Share election public key with another guardian
        :return: Election public key
        """
        return ElectionPublicKey(
            self.object_id,
            self._election_keys.proof,
            self._election_keys.key_pair.public_key,
        )

    def share_coefficient_validation_set(self) -> CoefficientValidationSet:
        """
        Share coefficient validation set to be used for validating the coefficients post election
        """
        return get_coefficient_validation_set(
            self.object_id, self._election_keys.polynomial
        )

    def save_election_public_key(self, key: ElectionPublicKey) -> None:
        """
        Save a guardians election public key
        :param key: Election public key
        """
        self._guardian_election_public_keys.set(key.owner_id, key)

    def all_election_public_keys_received(self) -> bool:
        """
        True if all election public keys have been received.
        :return: All election public keys backups received
        """
        return (
            len(self._guardian_election_public_keys)
            == self.ceremony_details.number_of_guardians
        )

    def guardian_election_public_keys(
        self,
    ) -> ReadOnlyDataStore[GUARDIAN_ID, ElectionPublicKey]:
        """
        Get a read-only view of the Guardian Election Public Keys shared with this Guardian

        """
        return ReadOnlyDataStore(self._guardian_election_public_keys)

    def generate_election_partial_key_backups(
        self, encrypt: AuxiliaryEncrypt = rsa_encrypt
    ) -> bool:
        """
        Generate all election partial key backups based on existing public keys
        :param encrypt: Encryption function using auxiliary key
        """
        if not self.all_auxiliary_public_keys_received():
            log_warning(
                f"guardian; {self.object_id} could not generate election partial key backups: missing auxiliary keys"
            )
            return False
        for auxiliary_key in self._guardian_auxiliary_public_keys.values():
            backup = generate_election_partial_key_backup(
                self.object_id, self._election_keys.polynomial, auxiliary_key, encrypt
            )
            if backup is None:
                log_warning(
                    f"guardian; {self.object_id} could not generate election partial key backups: failed to encrypt"
                )
                return False
            self._backups_to_share.set(auxiliary_key.owner_id, backup)

        return True

    # Election Partial Key Backup
    def share_election_partial_key_backup(
        self, designated_id: GUARDIAN_ID
    ) -> Optional[ElectionPartialKeyBackup]:
        """
        Share election partial key backup with another guardian
        :param designated_id: Designated guardian
        :return: Election partial key backup or None
        """
        return self._backups_to_share.get(designated_id)

    def save_election_partial_key_backup(
        self, backup: ElectionPartialKeyBackup
    ) -> None:
        """
        Save election partial key backup from another guardian
        :param backup: Election partial key backup
        """
        self._guardian_election_partial_key_backups.set(backup.owner_id, backup)

    def all_election_partial_key_backups_received(self) -> bool:
        """
        True if all election partial key backups have been received.
        :return: All election partial key backups received
        """
        return (
            len(self._guardian_election_partial_key_backups)
            == self.ceremony_details.number_of_guardians - 1
        )

    # Verification
    def verify_election_partial_key_backup(
        self, guardian_id: GUARDIAN_ID, decrypt: AuxiliaryDecrypt = rsa_decrypt,
    ) -> Optional[ElectionPartialKeyVerification]:
        """
        Verify election partial key backup value is in polynomial
        :param guardian_id: Owner of backup to verify
        :param decrypt: 
        :return: Election partial key verification or None
        """
        backup = self._guardian_election_partial_key_backups.get(guardian_id)
        if backup is None:
            return None
        return verify_election_partial_key_backup(
            self.object_id, backup, self._auxiliary_keys, decrypt
        )

    def publish_election_backup_challenge(
        self, guardian_id: GUARDIAN_ID
    ) -> Optional[ElectionPartialKeyChallenge]:
        """
        Publish election backup challenge of election partial key verification
        :param guardian_id: Owner of election key 
        :return: Election partial key challenge or None
        """
        backup_in_question = self._backups_to_share.get(guardian_id)
        if backup_in_question is None:
            return None
        return generate_election_partial_key_challenge(
            backup_in_question, self._election_keys.polynomial
        )

    def verify_election_partial_key_challenge(
        self, challenge: ElectionPartialKeyChallenge
    ) -> ElectionPartialKeyVerification:
        """
        Verify challenge of previous verification of election partial key
        :param challenge: Election partial key challenge
        :return: Election partial key verification 
        """
        return verify_election_partial_key_challenge(self.object_id, challenge)

    def save_election_partial_key_verification(
        self, verification: ElectionPartialKeyVerification
    ) -> None:
        """
        Save election partial key verification from another guardian
        :param verification: Election partial key verification
        """
        self._guardian_election_partial_key_verifications.set(
            verification.designated_id, verification
        )

    def all_election_partial_key_backups_verified(self) -> bool:
        """
        True if all election partial key backups have been verified.
        :return: All election partial key backups verified
        """
        required = self.ceremony_details.number_of_guardians - 1
        if len(self._guardian_election_partial_key_verifications) != required:
            return False
        for verified in self._guardian_election_partial_key_verifications.values():
            if not verified:
                return False
        return True

    # Joint Key
    def publish_joint_key(self) -> Optional[ElectionJointKey]:
        """
        Creates a joint election key from the public keys of all guardians
        :return: Optional joint key for election
        """
        if not self.all_election_public_keys_received():
            return None
        if not self.all_election_partial_key_backups_verified():
            return None
        return combine_election_public_keys(self._guardian_election_public_keys)

    def partially_decrypt(
        self,
        elgamal: ElGamalCiphertext,
        extended_base_hash: ElementModQ,
        nonce_seed: ElementModQ = None,
    ) -> Tuple[ElementModP, ChaumPedersenProof]:
        """
        Compute a partial decryption of an elgamal encryption

        :param elgamal: the `ElGamalCiphertext` that will be partially decrypted
        :param extended_base_hash: the extended base hash of the election that 
                                   was used to generate t he ElGamal Ciphertext
        :param nonce_seed: an optional value used to generate the `ChaumPedersenProof`
                           if no value is provided, a random number will be used.
        :return: a `Tuple[ElementModP, ChaumPedersenProof]` of the decryption and its proof
        """
        if nonce_seed is None:
            nonce_seed = rand_q()

        # TODO: ISSUE #47: Decrypt the election secret key

        # 𝑀_i = 𝐴^𝑠𝑖 mod 𝑝
        partial_decryption = elgamal.partial_decrypt(
            self._election_keys.key_pair.secret_key
        )

        # 𝑀_i = 𝐴^𝑠𝑖 mod 𝑝 and 𝐾𝑖 = 𝑔^𝑠𝑖 mod 𝑝
        proof = make_chaum_pedersen(
            message=elgamal,
            s=self._election_keys.key_pair.secret_key,
            m=partial_decryption,
            seed=nonce_seed,
            hash_header=extended_base_hash,
        )

        return (partial_decryption, proof)

    def compensate_decrypt(
        self,
        missing_guardian_id: str,
        elgamal: ElGamalCiphertext,
        extended_base_hash: ElementModQ,
        nonce_seed: ElementModQ = None,
        decrypt: AuxiliaryDecrypt = rsa_decrypt,
    ) -> Optional[Tuple[ElementModP, ChaumPedersenProof]]:
        """
        Compute a compensated partial decryption of an elgamal encryption 
        on behalf of the missing guardian

        :param missing_guardian_id: the guardian 
        :param elgamal: the `ElGamalCiphertext` that will be partially decrypted
        :param extended_base_hash: the extended base hash of the election that 
                                   was used to generate t he ElGamal Ciphertext
        :param nonce_seed: an optional value used to generate the `ChaumPedersenProof`
                           if no value is provided, a random number will be used.
        :param decrypt: an `AuxiliaryDecrypt` function to decrypt the missing guardina private key backup
        :return: a `Tuple[ElementModP, ChaumPedersenProof]` of the decryption and its proof
        """
        if nonce_seed is None:
            nonce_seed = rand_q()

        backup = self._guardian_election_partial_key_backups.get(missing_guardian_id)
        if backup is None:
            log_warning(
                f"compensate decrypt guardian {self.object_id} missing backup for {missing_guardian_id}"
            )
            return None

        decrypted_value = decrypt(
            backup.encrypted_value, self._auxiliary_keys.secret_key
        )
        if decrypted_value is None:
            log_warning(
                f"compensate decrypt guardian {self.object_id} failed decryption for {missing_guardian_id}"
            )
            return None
        partial_secret_key = get_optional(int_to_q(int(decrypted_value)))

        # 𝑀_{𝑖,l} = 𝐴^P𝑖_{l}
        partial_decryption = elgamal.partial_decrypt(partial_secret_key)

        # 𝑀_{𝑖,l} = 𝐴^𝑠𝑖 mod 𝑝 and 𝐾𝑖 = 𝑔^𝑠𝑖 mod 𝑝
        proof = make_chaum_pedersen(
            message=elgamal,
            s=partial_secret_key,
            m=partial_decryption,
            seed=nonce_seed,
            hash_header=extended_base_hash,
        )

        return (partial_decryption, proof)

    def recovery_public_key_for(
        self, missing_guardian_id: GUARDIAN_ID
    ) -> Optional[ElementModP]:
        """
        Compute the recovery public key for a given guardian
        """
        backup = self._guardian_election_partial_key_backups.get(missing_guardian_id)
        if backup is None:
            log_warning(
                f"compensate decrypt guardian {self.object_id} missing backup for {missing_guardian_id}"
            )
            return None

        # compute the recovery public key,
        # corresponding to the secret share Pi(l)
        # K_ij^(l^j) for j in 0..k-1.  K_ij is coefficients[j].public_key
        pub_key = ONE_MOD_P
        for index, commitment in enumerate(backup.coefficient_commitments):
            exponent = pow_q(self.sequence_order, index)
            pub_key = mult_p(pub_key, pow_p(commitment, exponent))

        return pub_key



from typing import Dict, List, Optional

from .auxiliary import AuxiliaryDecrypt
from .ballot import (
    CiphertextAcceptedBallot,
    CiphertextSelection,
)
from .data_store import DataStore
from .decryption_share import (
    BallotDecryptionShare,
    CompensatedBallotDecryptionShare,
    CiphertextDecryptionSelection,
    CiphertextCompensatedDecryptionSelection,
    CiphertextDecryptionContest,
    CiphertextCompensatedDecryptionContest,
    TallyDecryptionShare,
    CompensatedTallyDecryptionShare,
    create_ciphertext_decryption_selection,
)
from .election import CiphertextElectionContext
from .election_polynomial import compute_lagrange_coefficient
from .group import ElementModP, ElementModQ, mult_p, pow_p
from .guardian import Guardian, PublicKeySet
from .key_ceremony import ElectionPublicKey
from .logs import log_warning
from .rsa import rsa_decrypt
from .scheduler import Scheduler
from .tally import (
    CiphertextTally,
    CiphertextTallyContest,
)

from .types import BALLOT_ID, CONTEST_ID, GUARDIAN_ID, SELECTION_ID

AVAILABLE_GUARDIAN_ID = GUARDIAN_ID
MISSING_GUARDIAN_ID = GUARDIAN_ID

GUARDIAN_PUBLIC_KEY = ElementModP
ELECTION_PUBLIC_KEY = ElementModP


def compute_decryption_share(
    guardian: Guardian, tally: CiphertextTally, context: CiphertextElectionContext,
) -> Optional[TallyDecryptionShare]:
    """
    Compute a decryptions share for a guardian

    :param guardian: The guardian who will partially decrypt the tally
    :param tally: The election tally to decrypt
    :context: The public election encryption context
    :return: a `TallyDecryptionShare` or `None` if there is an error
    """

    contests = compute_decryption_share_for_cast_contests(guardian, tally, context)
    if contests is None:
        return None

    spoiled_ballots = compute_decryption_share_for_spoiled_ballots(
        guardian, tally, context
    )

    if spoiled_ballots is None:
        return None

    return TallyDecryptionShare(
        guardian.object_id,
        guardian.share_election_public_key().key,
        contests,
        spoiled_ballots,
    )


def compute_compensated_decryption_share(
    guardian: Guardian,
    missing_guardian_id: str,
    tally: CiphertextTally,
    context: CiphertextElectionContext,
    decrypt: AuxiliaryDecrypt = rsa_decrypt,
) -> Optional[CompensatedTallyDecryptionShare]:
    """
    Compute a compensated decryptions share for a guardian

    :param guardian: The guardian who will partially decrypt the tally
    :param missing_guardian_id: the missing guardian id to compensate
    :param tally: The election tally to decrypt
    :context: The public election encryption context
    :return: a `TallyDecryptionShare` or `None` if there is an error
    """

    contests = compute_compensated_decryption_share_for_cast_contests(
        guardian, missing_guardian_id, tally, context, decrypt
    )
    if contests is None:
        return None

    spoiled_ballots = compute_compensated_decryption_share_for_spoiled_ballots(
        guardian, missing_guardian_id, tally, context, decrypt
    )

    if spoiled_ballots is None:
        return None

    return CompensatedTallyDecryptionShare(
        guardian.object_id,
        missing_guardian_id,
        guardian.share_election_public_key().key,
        contests,
        spoiled_ballots,
    )


def compute_decryption_share_for_cast_contests(
    guardian: Guardian, tally: CiphertextTally, context: CiphertextElectionContext,
) -> Optional[Dict[CONTEST_ID, CiphertextDecryptionContest]]:
    """
    Compute the decryption for all of the cast contests in the Ciphertext Tally
    """
    contests: Dict[CONTEST_ID, CiphertextDecryptionContest] = {}
    scheduler = Scheduler()

    for contest in tally.cast.values():
        selections: Dict[SELECTION_ID, CiphertextDecryptionSelection] = {}
        selection_decryptions: List[
            Optional[CiphertextDecryptionSelection]
        ] = scheduler.schedule(
            compute_decryption_share_for_selection,
            [
                (guardian, selection, context)
                for (_, selection) in contest.tally_selections.items()
            ],
            with_shared_resources=True,
        )

        # verify the decryptions are received and add them to the collection
        for decryption in selection_decryptions:
            if decryption is None:
                log_warning(
                    f"could not compute share for guardian {guardian.object_id} contest {contest.object_id}"
                )
                return None
            selections[decryption.object_id] = decryption

        contests[contest.object_id] = CiphertextDecryptionContest(
            contest.object_id, guardian.object_id, contest.description_hash, selections
        )

    return contests


def compute_compensated_decryption_share_for_cast_contests(
    guardian: Guardian,
    missing_guardian_id: str,
    tally: CiphertextTally,
    context: CiphertextElectionContext,
    decrypt: AuxiliaryDecrypt = rsa_decrypt,
) -> Optional[Dict[CONTEST_ID, CiphertextCompensatedDecryptionContest]]:
    """
    Compute the compensated decryption for all of the cast contests in the Ciphertext Tally
    """
    scheduler = Scheduler()
    contests: Dict[CONTEST_ID, CiphertextCompensatedDecryptionContest] = {}

    for contest in tally.cast.values():
        selections: Dict[SELECTION_ID, CiphertextCompensatedDecryptionSelection] = {}
        selection_decryptions: List[
            Optional[CiphertextCompensatedDecryptionSelection]
        ] = scheduler.schedule(
            compute_compensated_decryption_share_for_selection,
            [
                (guardian, missing_guardian_id, selection, context, decrypt)
                for (_, selection) in contest.tally_selections.items()
            ],
            with_shared_resources=True,
        )

        # verify the decryptions are received and add them to the collection
        for decryption in selection_decryptions:
            if decryption is None:
                log_warning(
                    f"could not compute share for guardian {guardian.object_id} contest {contest.object_id}"
                )
                return None
            selections[decryption.object_id] = decryption

        contests[contest.object_id] = CiphertextCompensatedDecryptionContest(
            contest.object_id,
            guardian.object_id,
            missing_guardian_id,
            contest.description_hash,
            selections,
        )
    return contests


def compute_decryption_share_for_spoiled_ballots(
    guardian: Guardian, tally: CiphertextTally, context: CiphertextElectionContext,
) -> Optional[Dict[BALLOT_ID, BallotDecryptionShare]]:
    """
    Compute the decryption for all spoiled ballots in the Ciphertext Tally
    """
    spoiled_ballots: Dict[BALLOT_ID, BallotDecryptionShare] = {}
    scheduler = Scheduler()

    for spoiled_ballot in tally.spoiled_ballots.values():
        computed_share = compute_decryption_share_for_ballot(
            guardian, spoiled_ballot, context, scheduler
        )

        if computed_share:
            spoiled_ballots[spoiled_ballot.object_id] = computed_share
        else:
            return None

    return spoiled_ballots


def compute_decryption_share_for_ballot(
    guardian: Guardian,
    ballot: CiphertextAcceptedBallot,
    context: CiphertextElectionContext,
    scheduler: Optional[Scheduler] = None,
) -> Optional[BallotDecryptionShare]:
    """
    Compute the decryption for a single ballot
    """
    if not scheduler:
        scheduler = Scheduler()

    contests: Dict[CONTEST_ID, CiphertextDecryptionContest] = {}
    for contest in ballot.contests:
        selections: Dict[SELECTION_ID, CiphertextDecryptionSelection] = {}
        selection_decryptions: List[
            Optional[CiphertextDecryptionSelection]
        ] = scheduler.schedule(
            compute_decryption_share_for_selection,
            [(guardian, selection, context) for selection in contest.ballot_selections],
            with_shared_resources=True,
        )
        # verify the decryptions are received and add them to the collection
        for decryption in selection_decryptions:
            if decryption is None:
                log_warning(
                    f"could not compute spoiled ballot share for guardian {guardian.object_id} contest {contest.object_id}"
                )
                return None
            selections[decryption.object_id] = decryption

        contests[contest.object_id] = CiphertextDecryptionContest(
            contest.object_id, guardian.object_id, contest.description_hash, selections,
        )
    return BallotDecryptionShare(
        guardian.object_id,
        guardian.share_election_public_key().key,
        ballot.object_id,
        contests,
    )


def compute_compensated_decryption_share_for_spoiled_ballots(
    guardian: Guardian,
    missing_guardian_id: MISSING_GUARDIAN_ID,
    tally: CiphertextTally,
    context: CiphertextElectionContext,
    decrypt: AuxiliaryDecrypt = rsa_decrypt,
) -> Optional[Dict[BALLOT_ID, CompensatedBallotDecryptionShare]]:
    """
    Compute the decryption for all spoiled ballots in the Ciphertext Tally
    """
    spoiled_ballots: Dict[BALLOT_ID, CompensatedBallotDecryptionShare] = {}
    scheduler = Scheduler()

    for spoiled_ballot in tally.spoiled_ballots.values():
        contests: Dict[CONTEST_ID, CiphertextCompensatedDecryptionContest] = {}
        compensated_ballot = compute_compensated_decryption_share_for_ballot(
            guardian, missing_guardian_id, spoiled_ballot, context, decrypt, scheduler
        )

        if compensated_ballot:
            spoiled_ballots[spoiled_ballot.object_id] = compensated_ballot
        else:
            return None
    return spoiled_ballots


def compute_compensated_decryption_share_for_ballot(
    guardian: Guardian,
    missing_guardian_id: MISSING_GUARDIAN_ID,
    ballot: CiphertextAcceptedBallot,
    context: CiphertextElectionContext,
    decrypt: AuxiliaryDecrypt = rsa_decrypt,
    scheduler: Optional[Scheduler] = None,
) -> Optional[CompensatedBallotDecryptionShare]:
    """
    Compute the decryption for a single ballot
    """
    if not scheduler:
        scheduler = Scheduler()

    contests: Dict[CONTEST_ID, CiphertextCompensatedDecryptionContest] = {}
    for contest in ballot.contests:
        selections: Dict[SELECTION_ID, CiphertextCompensatedDecryptionSelection] = {}
        selection_decryptions: List[
            Optional[CiphertextCompensatedDecryptionSelection]
        ] = scheduler.schedule(
            compute_compensated_decryption_share_for_selection,
            [
                (guardian, missing_guardian_id, selection, context, decrypt)
                for selection in contest.ballot_selections
            ],
            with_shared_resources=True,
        )
        # verify the decryptions are received and add them to the collection
        for decryption in selection_decryptions:
            if decryption is None:
                log_warning(
                    f"could not compute compensated spoiled ballot share for guardian {guardian.object_id} missing: {missing_guardian_id} contest {contest.object_id}"
                )
                return None
            selections[decryption.object_id] = decryption

        contests[contest.object_id] = CiphertextCompensatedDecryptionContest(
            contest.object_id,
            guardian.object_id,
            missing_guardian_id,
            contest.description_hash,
            selections,
        )
    return CompensatedBallotDecryptionShare(
        guardian.object_id,
        missing_guardian_id,
        guardian.share_election_public_key().key,
        ballot.object_id,
        contests,
    )


def compute_decryption_share_for_selection(
    guardian: Guardian,
    selection: CiphertextSelection,
    context: CiphertextElectionContext,
) -> Optional[CiphertextDecryptionSelection]:
    """
    Compute a partial decryption for a specific selection

    :param guardian: The guardian who will partially decrypt the selection
    :param selection: The specific selection to decrypt
    :context: The public election encryption context
    :return: a `CiphertextDecryptionSelection` or `None` if there is an error
    """

    (decryption, proof) = guardian.partially_decrypt(
        selection.ciphertext, context.crypto_extended_base_hash
    )

    if proof.is_valid(
        selection.ciphertext,
        guardian.share_election_public_key().key,
        decryption,
        context.crypto_extended_base_hash,
    ):
        return create_ciphertext_decryption_selection(
            selection.object_id,
            guardian.object_id,
            selection.description_hash,
            decryption,
            proof,
        )
    else:
        log_warning(
            f"compute decryption share proof failed for {guardian.object_id} {selection.object_id} with invalid proof"
        )
        return None


def compute_compensated_decryption_share_for_selection(
    available_guardian: Guardian,
    missing_guardian_id: str,
    selection: CiphertextSelection,
    context: CiphertextElectionContext,
    decrypt: AuxiliaryDecrypt = rsa_decrypt,
) -> Optional[CiphertextCompensatedDecryptionSelection]:
    """
    Compute a compensated decryption share for a specific selection using the 
    avialable guardian's share of the missing guardian's private key polynomial

    :param available_guardian: The available guardian that will partially decrypt the selection
    :param missing_guardian_id: The id of the guardian that is missing
    :param selection: The specific selection to decrypt
    :context: The public election encryption context
    :return: a `CiphertextCompensatedDecryptionSelection` or `None` if there is an error
    """

    compensated = available_guardian.compensate_decrypt(
        missing_guardian_id,
        selection.ciphertext,
        context.crypto_extended_base_hash,
        decrypt=decrypt,
    )

    if compensated is None:
        log_warning(
            f"compute compensated decryption share failed for {available_guardian.object_id} missing: {missing_guardian_id} {selection.object_id}"
        )
        return None

    (decryption, proof) = compensated

    recovery_public_key = available_guardian.recovery_public_key_for(
        missing_guardian_id
    )

    if recovery_public_key is None:
        log_warning(
            f"compute compensated decryption share failed for {available_guardian.object_id} missing recovery key: {missing_guardian_id} {selection.object_id}"
        )
        return None

    if proof.is_valid(
        selection.ciphertext,
        recovery_public_key,
        decryption,
        context.crypto_extended_base_hash,
    ):
        share = CiphertextCompensatedDecryptionSelection(
            selection.object_id,
            available_guardian.object_id,
            missing_guardian_id,
            selection.description_hash,
            decryption,
            recovery_public_key,
            proof,
        )
        return share
    else:
        log_warning(
            f"compute compensated decryption share proof failed for {available_guardian.object_id} missing: {missing_guardian_id} {selection.object_id}"
        )
        return None


def compute_lagrange_coefficients_for_guardian(
    all_available_guardian_keys: List[PublicKeySet], guardian_keys: PublicKeySet
) -> ElementModQ:
    """
    Produce a Lagrange coefficient for a single Guardian, to be used when reconstructing a missing share.
    """
    other_guardian_orders = [
        g.sequence_order
        for g in all_available_guardian_keys
        if g.owner_id != guardian_keys.owner_id
    ]
    return compute_lagrange_coefficient(
        guardian_keys.sequence_order, *other_guardian_orders,
    )


def compute_lagrange_coefficients_for_guardians(
    all_available_guardian_keys: List[PublicKeySet],
) -> Dict[AVAILABLE_GUARDIAN_ID, ElementModQ]:
    """
    Produce all Lagrange coefficients for a collection of available Guardians, to be used when reconstructing a missing share.
    """
    return {
        guardian_keys.owner_id: compute_lagrange_coefficients_for_guardian(
            all_available_guardian_keys, guardian_keys
        )
        for guardian_keys in all_available_guardian_keys
    }


def reconstruct_missing_tally_decryption_shares(
    ciphertext_tally: CiphertextTally,
    missing_guardians: DataStore[MISSING_GUARDIAN_ID, ElectionPublicKey],
    compensated_shares: Dict[
        MISSING_GUARDIAN_ID,
        Dict[AVAILABLE_GUARDIAN_ID, CompensatedTallyDecryptionShare],
    ],
    lagrange_coefficients: Dict[
        MISSING_GUARDIAN_ID, Dict[AVAILABLE_GUARDIAN_ID, ElementModQ]
    ],
) -> Optional[Dict[MISSING_GUARDIAN_ID, TallyDecryptionShare]]:
    """
    Use all available guardians to reconstruct the missing shares for all missing guardians
    """

    reconstructed_shares: Dict[MISSING_GUARDIAN_ID, TallyDecryptionShare] = {}
    for missing_guardian_id, shares in compensated_shares.items():

        # Make sure there is a public key
        public_key = missing_guardians.get(missing_guardian_id)
        if public_key is None:
            log_warning(
                f"Could not reconstruct tally for {missing_guardian_id} with no public key"
            )
            return None

        # make sure there are computed lagrange coefficients:
        lagrange_coefficients_for_missing: Dict[
            AVAILABLE_GUARDIAN_ID, ElementModQ
        ] = lagrange_coefficients.get(missing_guardian_id, {})
        if not any(lagrange_coefficients):
            log_warning(
                f"Could not reconstruct tally for {missing_guardian_id} with no lagrange coefficients"
            )
            return None

        # iterate through the tallies and accumulate
        # all of the shares for this guardian
        contests = reconstruct_decryption_contests(
            missing_guardian_id,
            ciphertext_tally.cast,
            shares,
            lagrange_coefficients_for_missing,
        )

        # iterate through the spoiled ballots and accumulate
        # all of the shares for this guardian
        spoiled_ballots = reconstruct_decryption_ballots(
            missing_guardian_id,
            public_key,
            ciphertext_tally.spoiled_ballots,
            shares,
            lagrange_coefficients_for_missing,
        )

        reconstructed_shares[missing_guardian_id] = TallyDecryptionShare(
            missing_guardian_id, public_key.key, contests, spoiled_ballots
        )

    return reconstructed_shares


def reconstruct_decryption_contests(
    missing_guardian_id: MISSING_GUARDIAN_ID,
    cast_tally: Dict[CONTEST_ID, CiphertextTallyContest],
    shares: Dict[AVAILABLE_GUARDIAN_ID, CompensatedTallyDecryptionShare],
    lagrange_coefficients: Dict[AVAILABLE_GUARDIAN_ID, ElementModQ],
) -> Dict[CONTEST_ID, CiphertextDecryptionContest]:
    """
    Recontruct the missing Decryption Share for a missing guardian 
    from the collection of compensated decryption shares

    :param missing_guardian_id: The guardian id for the missing guardian
    :param cast_tally: The collection of `CiphertextTallyContest` that is cast
    :shares: the collection of `CompensatedTallyDecryptionShare` for the missing guardian
    :lagrange_coefficients: the lagrange coefficients corresponding to the available guardians that provided shares
    """
    # iterate through the tallies and accumulate all of the shares for this guardian
    contests: Dict[CONTEST_ID, CiphertextDecryptionContest] = {}
    for contest_id, tally_contest in cast_tally.items():
        contest_shares: Dict[
            AVAILABLE_GUARDIAN_ID, CiphertextCompensatedDecryptionContest
        ] = {
            available_guardian_id: compensated_share.contests[tally_contest.object_id]
            for available_guardian_id, compensated_share in shares.items()
        }

        selections: Dict[SELECTION_ID, CiphertextDecryptionSelection] = {}
        for (selection_id, tally_selection,) in tally_contest.tally_selections.items():

            # collect all of the shares generated for each selection
            compensated_selection_shares: Dict[
                AVAILABLE_GUARDIAN_ID, CiphertextCompensatedDecryptionSelection
            ] = {
                available_guardian_id: compensated_contest.selections[selection_id]
                for available_guardian_id, compensated_contest in contest_shares.items()
            }

            share_pow_p = []
            for available_guardian_id, share in compensated_selection_shares.items():
                share_pow_p.append(
                    pow_p(share.share, lagrange_coefficients[available_guardian_id])
                )

            reconstructed_share = mult_p(*share_pow_p)

            selections[selection_id] = create_ciphertext_decryption_selection(
                selection_id,
                missing_guardian_id,
                tally_selection.description_hash,
                reconstructed_share,
                compensated_selection_shares,
            )
        contests[contest_id] = CiphertextDecryptionContest(
            contest_id, missing_guardian_id, tally_contest.description_hash, selections,
        )

    return contests


def reconstruct_decryption_ballots(
    missing_guardian_id: MISSING_GUARDIAN_ID,
    public_key: ElectionPublicKey,
    spoiled_ballots: Dict[BALLOT_ID, CiphertextAcceptedBallot],
    shares: Dict[AVAILABLE_GUARDIAN_ID, CompensatedTallyDecryptionShare],
    lagrange_coefficients: Dict[AVAILABLE_GUARDIAN_ID, ElementModQ],
) -> Dict[BALLOT_ID, BallotDecryptionShare]:
    """
    Recontruct the missing Decryption shares for a missing guardian from the collection of compensated decryption shares

    :param missing_guardian_id: The guardian id for the missing guardian
    :param public_key: the public key for the missing guardian
    :param spoiled_ballots: The collection of `CiphertextAcceptedBallot` that is spoiled
    :shares: the collection of `CompensatedTallyDecryptionShare` for the missing guardian
    :lagrange_coefficients: the lagrange coefficients corresponding to the available guardians that provided shares
    """
    spoiled_ballot_shares: Dict[BALLOT_ID, BallotDecryptionShare] = {}
    for ballot_id, spoiled_ballot in spoiled_ballots.items():
        # iterate through the tallies and accumulate all of the shares for this guardian
        shares_for_ballot: Dict[GUARDIAN_ID, CompensatedBallotDecryptionShare] = {
            available_guardian_id: compensated_ballot
            for available_guardian_id, compensated_share in shares.items()
            for compensated_ballot_id, compensated_ballot in compensated_share.spoiled_ballots.items()
            if compensated_ballot_id == ballot_id
        }
        ballot_share = reconstruct_decryption_ballot(
            missing_guardian_id,
            public_key,
            spoiled_ballot,
            shares_for_ballot,
            lagrange_coefficients,
        )

        spoiled_ballot_shares[spoiled_ballot.object_id] = ballot_share
    return spoiled_ballot_shares


def reconstruct_decryption_ballot(
    missing_guardian_id: MISSING_GUARDIAN_ID,
    public_key: ElectionPublicKey,
    ballot: CiphertextAcceptedBallot,
    shares: Dict[AVAILABLE_GUARDIAN_ID, CompensatedBallotDecryptionShare],
    lagrange_coefficients: Dict[AVAILABLE_GUARDIAN_ID, ElementModQ],
) -> BallotDecryptionShare:
    """
    Reconstruct a missing ballot Decryption share for a missing guardian from the collection of compensated decryption shares

    :param missing_guardian_id: The guardian id for the missing guardian
    :param public_key: the public key for the missing guardian
    :param ballots: The `CiphertextAcceptedBallot` to reconstruct
    :shares: the collection of `CompensatedBallotDecryptionShare` for the missing guardian, each keyed by the ID of the guardian that produced it
    :lagrange_coefficients: the lagrange coefficients corresponding to the available guardians that provided shares
    """
    contests: Dict[CONTEST_ID, CiphertextDecryptionContest] = {}
    for contest in ballot.contests:
        contest_shares: Dict[
            AVAILABLE_GUARDIAN_ID, CiphertextCompensatedDecryptionContest
        ] = {
            available_guardian_id: compensated_ballot.contests[contest.object_id]
            for available_guardian_id, compensated_ballot in shares.items()
        }

        selections: Dict[SELECTION_ID, CiphertextDecryptionSelection] = {}
        for selection in contest.ballot_selections:

            # collect all of the shares generated for each selection
            compensated_selection_shares: Dict[
                AVAILABLE_GUARDIAN_ID, CiphertextCompensatedDecryptionSelection
            ] = {
                available_guardian_id: compensated_contest.selections[
                    selection.object_id
                ]
                for available_guardian_id, compensated_contest in contest_shares.items()
            }

            # compute the reconstructed share
            reconstructed_share = mult_p(
                *[
                    pow_p(share.share, lagrange_coefficients[available_guardian_id])
                    for available_guardian_id, share in compensated_selection_shares.items()
                ]
            )
            selections[selection.object_id] = create_ciphertext_decryption_selection(
                selection.object_id,
                missing_guardian_id,
                selection.description_hash,
                reconstructed_share,
                compensated_selection_shares,
            )
        contests[contest.object_id] = CiphertextDecryptionContest(
            contest.object_id,
            missing_guardian_id,
            contest.description_hash,
            selections,
        )
    return BallotDecryptionShare(
        missing_guardian_id, public_key.key, ballot.object_id, contests,
    )



from typing import List, Optional

from .hash import hash_elems
from .group import ElementModQ, q_to_bytes, bytes_to_q
from .words import get_word, get_index_from_word

DEFAULT_SEPERATOR = "-"


def get_hash_for_device(uuid: int, location: str) -> ElementModQ:
    """
    Get starting hash for given device
    :param uuid: Unique identifier of device
    :param location: Location of device
    :return: Starting hash of device
    """
    return hash_elems(uuid, location)


def get_rotating_tracker_hash(
    prev_hash: ElementModQ, timestamp: int, ballot_hash: ElementModQ
) -> ElementModQ:
    """
    Get the rotated tracker hash for a particular ballot. 
    :param prev_hash: Previous hash or starting hash from device
    :param timestamp: Timestamp in ticks
    :param ballot_hash: Hash of ballot to track
    :return: Tracker hash
    """
    return hash_elems(prev_hash, timestamp, ballot_hash)


def tracker_hash_to_words(
    tracker_hash: ElementModQ, seperator: str = DEFAULT_SEPERATOR
) -> Optional[str]:
    """
    Convert tracker has to human readable / friendly words
    :param hash: Tracker hash
    :return: Human readable tracker string or None
    """

    segments = q_to_bytes(tracker_hash)
    words: List[str] = []
    for value in segments:
        word = get_word(value)
        if word is None:
            return None
        words.append(word)
    # FIXME ISSUE #82 Minimize length of tracker
    return seperator.join(words)


def tracker_words_to_hash(
    tracker_words: str, seperator: str = DEFAULT_SEPERATOR
) -> Optional[ElementModQ]:
    """
    Convert tracker from human readable / friendly words to hash
    :param tracker_words: Tracker words
    :param seperator: Seperator used between words
    :return: Tracker hash or None
    """
    words = tracker_words.split(seperator)
    int_values: List[int] = []
    for word in words:
        index = get_index_from_word(word)
        if index is None:
            return None
        int_values.append(index)
    value = bytes(int_values)
    return bytes_to_q(value)



from datetime import datetime
import os
from dataclasses import dataclass
from typing import TypeVar, Callable, Optional, Tuple, List

from hypothesis.strategies import (
    composite,
    emails,
    integers,
    text,
    SearchStrategy,
)

from electionguard.ballot import PlaintextBallot

from electionguard.election import (
    BallotStyle,
    CiphertextElectionContext,
    ElectionConstants,
    ElectionDescription,
    ElectionType,
    InternalElectionDescription,
    generate_placeholder_selections_from,
    GeopoliticalUnit,
    Candidate,
    Party,
    ContestDescription,
    SelectionDescription,
    ReportingUnitType,
    VoteVariationType,
    contest_description_with_placeholders_from,
    CandidateContestDescription,
    ReferendumContestDescription,
)

from electionguard.election_builder import ElectionBuilder

from electionguard.encrypt import contest_from

from electionguard.group import ElementModP
from electionguard.guardian import Guardian
from electionguard.key_ceremony import CoefficientValidationSet
from electionguard.key_ceremony_mediator import KeyCeremonyMediator
from electionguard.utils import get_optional

_T = TypeVar("_T")
_DrawType = Callable[[SearchStrategy[_T]], _T]

data = os.path.realpath(os.path.join(__file__, "../../../data"))

NUMBER_OF_GUARDIANS = 5
QUORUM = 3


@dataclass
class AllPublicElectionData:
    description: ElectionDescription
    metadata: InternalElectionDescription
    context: CiphertextElectionContext
    constants: ElectionConstants
    guardians: List[CoefficientValidationSet]


@dataclass
class AllPrivateElectionData:
    guardians: List[Guardian]


class ElectionFactory(object):

    simple_election_manifest_file_name = "election_manifest_simple.json"

    def get_simple_election_from_file(self) -> ElectionDescription:
        return self._get_election_from_file(self.simple_election_manifest_file_name)

    def get_hamilton_election_from_file(self) -> ElectionDescription:
        with open(
            os.path.join(data, "hamilton-county", "election_manifest.json"), "r"
        ) as subject:
            result = subject.read()
            target = ElectionDescription.from_json(result)

        return target

    def get_hamilton_election_with_encryption_context(
        self,
    ) -> Tuple[AllPublicElectionData, AllPrivateElectionData]:
        guardians: List[Guardian] = []
        coefficient_validation_sets: List[CoefficientValidationSet] = []

        # Configure the election builder
        description = self.get_hamilton_election_from_file()
        builder = ElectionBuilder(NUMBER_OF_GUARDIANS, QUORUM, description)

        # Setup Guardians
        for i in range(NUMBER_OF_GUARDIANS):
            guardians.append(
                Guardian(
                    "hamilton-county-canvass-board-member-" + str(i),
                    i,
                    NUMBER_OF_GUARDIANS,
                    QUORUM,
                )
            )

        # Run the key ceremony
        mediator = KeyCeremonyMediator(guardians[0].ceremony_details)
        for guardian in guardians:
            mediator.announce(guardian)
        mediator.orchestrate()
        mediator.verify()

        # Joint Key
        joint_key = mediator.publish_joint_key()

        # Save Validation Keys
        for guardian in guardians:
            coefficient_validation_sets.append(
                guardian.share_coefficient_validation_set()
            )

        builder.set_public_key(get_optional(joint_key))
        metadata, context = get_optional(builder.build())
        constants = ElectionConstants()

        return (
            AllPublicElectionData(
                description, metadata, context, constants, coefficient_validation_sets
            ),
            AllPrivateElectionData(guardians),
        )

    def get_fake_election(self) -> ElectionDescription:
        """
        Get a single Fake Election object that is manually constructed with default values
        """

        fake_ballot_style = BallotStyle("some-ballot-style-id")
        fake_ballot_style.geopolitical_unit_ids = ["some-geopoltical-unit-id"]

        fake_referendum_ballot_selections = [
            # Referendum selections are simply a special case of `candidate` in the object model
            SelectionDescription(
                "some-object-id-affirmative", "some-candidate-id-1", 0
            ),
            SelectionDescription("some-object-id-negative", "some-candidate-id-2", 1),
        ]

        sequence_order = 0
        number_elected = 1
        votes_allowed = 1
        fake_referendum_contest = ReferendumContestDescription(
            "some-referendum-contest-object-id",
            "some-geopoltical-unit-id",
            sequence_order,
            VoteVariationType.one_of_m,
            number_elected,
            votes_allowed,
            "some-referendum-contest-name",
            fake_referendum_ballot_selections,
        )

        fake_candidate_ballot_selections = [
            SelectionDescription(
                "some-object-id-candidate-1", "some-candidate-id-1", 0
            ),
            SelectionDescription(
                "some-object-id-candidate-2", "some-candidate-id-2", 1
            ),
            SelectionDescription(
                "some-object-id-candidate-3", "some-candidate-id-3", 2
            ),
        ]

        sequence_order_2 = 1
        number_elected_2 = 2
        votes_allowed_2 = 2
        fake_candidate_contest = CandidateContestDescription(
            "some-candidate-contest-object-id",
            "some-geopoltical-unit-id",
            sequence_order_2,
            VoteVariationType.one_of_m,
            number_elected_2,
            votes_allowed_2,
            "some-candidate-contest-name",
            fake_candidate_ballot_selections,
        )

        fake_election = ElectionDescription(
            election_scope_id="some-scope-id",
            type=ElectionType.unknown,
            start_date=datetime.now(),
            end_date=datetime.now(),
            geopolitical_units=[
                GeopoliticalUnit(
                    "some-geopoltical-unit-id",
                    "some-gp-unit-name",
                    ReportingUnitType.unknown,
                )
            ],
            parties=[Party("some-party-id-1"), Party("some-party-id-2")],
            candidates=[
                Candidate("some-candidate-id-1"),
                Candidate("some-candidate-id-2"),
                Candidate("some-candidate-id-3"),
            ],
            contests=[fake_referendum_contest, fake_candidate_contest],
            ballot_styles=[fake_ballot_style],
        )

        return fake_election

    def get_fake_ciphertext_election(
        self, description: ElectionDescription, elgamal_public_key: ElementModP
    ) -> Tuple[InternalElectionDescription, CiphertextElectionContext]:
        builder = ElectionBuilder(
            number_of_guardians=1, quorum=1, description=description
        )
        builder.set_public_key(elgamal_public_key)
        metadata, election = get_optional(builder.build())
        return metadata, election

    # TODO: Move to ballot Factory?
    def get_fake_ballot(
        self, election: ElectionDescription = None, ballot_id: str = None
    ) -> PlaintextBallot:
        """
        Get a single Fake Ballot object that is manually constructed with default vaules
        """
        if election is None:
            election = self.get_fake_election()

        if ballot_id is None:
            ballot_id = "some-unique-ballot-id-123"

        fake_ballot = PlaintextBallot(
            ballot_id,
            election.ballot_styles[0].object_id,
            [contest_from(election.contests[0]), contest_from(election.contests[1])],
        )

        return fake_ballot

    def _get_election_from_file(self, filename: str) -> ElectionDescription:
        with open(os.path.join(data, filename), "r") as subject:
            result = subject.read()
            target = ElectionDescription.from_json(result)

        return target


@composite
def get_selection_description_well_formed(
    draw: _DrawType,
    ints=integers(1, 20),
    emails=emails(),
    candidate_id: Optional[str] = None,
    sequence_order: Optional[int] = None,
) -> Tuple[str, SelectionDescription]:
    if candidate_id is None:
        candidate_id = draw(emails)

    object_id = f"{candidate_id}-selection"

    if sequence_order is None:
        sequence_order = draw(ints)

    return (object_id, SelectionDescription(object_id, candidate_id, sequence_order))


@composite
def get_contest_description_well_formed(
    draw: _DrawType,
    ints=integers(1, 20),
    text=text(),
    emails=emails(),
    selections=get_selection_description_well_formed(),
    sequence_order: Optional[int] = None,
    electoral_district_id: Optional[str] = None,
) -> Tuple[str, ContestDescription]:
    object_id = f"{draw(emails)}-contest"

    if sequence_order is None:
        sequence_order = draw(ints)

    if electoral_district_id is None:
        electoral_district_id = f"{draw(emails)}-gp-unit"

    first_int = draw(ints)
    second_int = draw(ints)

    # TODO ISSUE #33: support more votes than seats for other VoteVariationType options
    number_elected = min(first_int, second_int)
    votes_allowed = number_elected

    selection_descriptions: List[SelectionDescription] = list()
    for i in range(max(first_int, second_int)):
        selection: Tuple[str, SelectionDescription] = draw(selections)
        _, selection_description = selection
        selection_description.sequence_order = i
        selection_descriptions.append(selection_description)

    contest_description = ContestDescription(
        object_id,
        electoral_district_id,
        sequence_order,
        VoteVariationType.n_of_m,
        number_elected,
        votes_allowed,
        draw(text),
        selection_descriptions,
    )

    placeholder_selections = generate_placeholder_selections_from(
        contest_description, number_elected
    )

    return (
        object_id,
        contest_description_with_placeholders_from(
            contest_description, placeholder_selections
        ),
    )



from collections.abc import Iterable

from typing import (
    Dict,
    Iterable,
    Iterator,
    Optional,
    List,
    Tuple,
)

from .ballot import BallotBoxState, CiphertextAcceptedBallot
from .logs import log_warning
from .types import BALLOT_ID

# TODO: ISSUE # 74: Remove this class in favor of using DataStore
class BallotStore(Iterable):
    """
    A representation of a cache of ballots for an election
    """

    _ballot_store: Dict[BALLOT_ID, Optional[CiphertextAcceptedBallot]]

    def __init__(self) -> None:
        self._ballot_store = {}

    def __iter__(self) -> Iterator[Optional[CiphertextAcceptedBallot]]:
        return iter(self._ballot_store.values())

    def set(
        self, ballot_id: str, ballot: Optional[CiphertextAcceptedBallot] = None
    ) -> bool:
        """
        Set a specific ballot id to a specific ballot
        """
        if ballot is not None and ballot.state == BallotBoxState.UNKNOWN:
            log_warning(f"cannot add ballot {ballot_id} to store with UNKNOWN state")
            return False
        self._ballot_store[ballot_id] = ballot
        return True

    def all(self) -> List[CiphertextAcceptedBallot]:
        """
        Get all `CiphertextAcceptedBallot` from the store
        """
        return [ballot for ballot in self._ballot_store.values() if ballot is not None]

    def get(self, ballot_id: str) -> Optional[CiphertextAcceptedBallot]:
        """
        Get a CiphertextAcceptedBallot from the store if it exists
        """
        if ballot_id in self._ballot_store:
            return self._ballot_store[ballot_id]

        return None

    def exists(self, ballot_id: str) -> Tuple[bool, Optional[CiphertextAcceptedBallot]]:
        """
        Check if a specific ballot exists and return it.
        :return: `Tuple[bool, Optional[CiphertextAcceptedBallot]]`
        """
        existing_ballot = self.get(ballot_id)
        if existing_ballot is None:
            return False, None
        return existing_ballot.state != BallotBoxState.UNKNOWN, existing_ballot



from typing import Iterable, NamedTuple, Optional

from .dlog import discrete_log
from .group import (
    ElementModQ,
    ElementModP,
    g_pow_p,
    mult_p,
    mult_inv_p,
    pow_p,
    ZERO_MOD_Q,
    TWO_MOD_Q,
    int_to_q,
    rand_range_q,
)
from .hash import hash_elems
from .logs import log_error
from .utils import flatmap_optional, get_optional


class ElGamalKeyPair(NamedTuple):
    """A tuple of an ElGamal secret key and public key."""

    secret_key: ElementModQ
    public_key: ElementModP


class ElGamalCiphertext(NamedTuple):
    """
    An "exponential ElGamal ciphertext" (i.e., with the plaintext in the exponent to allow for
    homomorphic addition). Create one with `elgamal_encrypt`. Add them with `elgamal_add`.
    Decrypt using one of the supplied instance methods.
    """

    pad: ElementModP
    """pad or alpha"""

    data: ElementModP
    """encrypted data or beta"""

    def decrypt_known_product(self, product: ElementModP) -> int:
        """
        Decrypts an ElGamal ciphertext with a "known product" (the blinding factor used in the encryption).

        :param product: The known product (blinding factor).
        :return: An exponentially encoded plaintext message.
        """
        return discrete_log(mult_p(self.data, mult_inv_p(product)))

    def decrypt(self, secret_key: ElementModQ) -> int:
        """
        Decrypt an ElGamal ciphertext using a known ElGamal secret key.

        :param secret_key: The corresponding ElGamal secret key.
        :return: An exponentially encoded plaintext message.
        """
        return self.decrypt_known_product(pow_p(self.pad, secret_key))

    def decrypt_known_nonce(self, public_key: ElementModP, nonce: ElementModQ) -> int:
        """
        Decrypt an ElGamal ciphertext using a known nonce and the ElGamal public key.

        :param public_key: The corresponding ElGamal public key.
        :param nonce: The secret nonce used to create the ciphertext.
        :return: An exponentially encoded plaintext message.
        """
        return self.decrypt_known_product(pow_p(public_key, nonce))

    def partial_decrypt(self, secret_key: ElementModQ) -> ElementModP:
        """
        Partially Decrypts an ElGamal ciphertext with a known ElGamal secret key.

        𝑀_i = 𝐴^𝑠𝑖 mod 𝑝 in the spec

        :param secret_key: The corresponding ElGamal secret key.
        :return: An exponentially encoded plaintext message.
        """
        return pow_p(self.pad, secret_key)

    def crypto_hash(self) -> ElementModQ:
        """
        Computes a cryptographic hash of this ciphertext.
        """
        return hash_elems(self.pad, self.data)


def elgamal_keypair_from_secret(a: ElementModQ) -> Optional[ElGamalKeyPair]:
    """
    Given an ElGamal secret key (typically, a random number in [2,Q)), returns
    an ElGamal keypair, consisting of the given secret key a and public key g^a.
    """
    secret_key_int = a.to_int()
    if secret_key_int < 2:
        log_error("ElGamal secret key needs to be in [2,Q).")
        return None

    return ElGamalKeyPair(a, g_pow_p(a))


def elgamal_keypair_random() -> ElGamalKeyPair:
    """
    Create a random elgamal keypair

    :return: random elgamal key pair
    """
    return get_optional(elgamal_keypair_from_secret(rand_range_q(TWO_MOD_Q)))


def elgamal_combine_public_keys(keys: Iterable[ElementModP]) -> ElementModP:
    """
    Combine multiple elgamal public keys into a joint key

    :param keys: list of public elgamal keys
    :return: joint key of elgamal keys
    """
    return mult_p(*keys)


def elgamal_encrypt(
    m: int, nonce: ElementModQ, public_key: ElementModP
) -> Optional[ElGamalCiphertext]:
    """
    Encrypts a message with a given random nonce and an ElGamal public key.

    :param m: Message to elgamal_encrypt; must be an integer in [0,Q).
    :param nonce: Randomly chosen nonce in [1,Q).
    :param public_key: ElGamal public key.
    :return: A ciphertext tuple.
    """
    if nonce == ZERO_MOD_Q:
        log_error("ElGamal encryption requires a non-zero nonce")
        return None

    return flatmap_optional(
        int_to_q(m),
        lambda e: ElGamalCiphertext(
            g_pow_p(nonce), mult_p(g_pow_p(e), pow_p(public_key, nonce))
        ),
    )


def elgamal_add(*ciphertexts: ElGamalCiphertext) -> ElGamalCiphertext:
    """
    Homomorphically accumulates one or more ElGamal ciphertexts by pairwise multiplication. The exponents
    of vote counters will add.
    """
    assert len(ciphertexts) != 0, "Must have one or more ciphertexts for elgamal_add"

    result = ciphertexts[0]
    for c in ciphertexts[1:]:
        result = ElGamalCiphertext(
            mult_p(result.pad, c.pad), mult_p(result.data, c.data)
        )

    return result



from typing import Callable, Optional, NamedTuple

from .types import GUARDIAN_ID

MESSAGE = str
PUBLIC_KEY = str
SECRET_KEY = str
ENCRYPTED_MESSAGE = str


class AuxiliaryKeyPair(NamedTuple):
    """A tuple of a secret key and public key."""

    secret_key: SECRET_KEY
    """The secret or private key"""
    public_key: PUBLIC_KEY


class AuxiliaryPublicKey(NamedTuple):
    """A tuple of auxiliary public key and owner information"""

    owner_id: GUARDIAN_ID
    """
    The unique identifier of the guardian
    """

    sequence_order: int
    """
    The sequence order of the auxiliary public key (usually the guardian's sequence order)
    """

    key: PUBLIC_KEY
    """
    A string representation of the Auxiliary public key.  
    It is up to the external `AuxiliaryEncrypt` function to know how to parse this value
    """


AuxiliaryEncrypt = Callable[[MESSAGE, PUBLIC_KEY], Optional[ENCRYPTED_MESSAGE]]
"""A callable type that represents the auxiliary encryption scheme."""

AuxiliaryDecrypt = Callable[[ENCRYPTED_MESSAGE, SECRET_KEY], Optional[MESSAGE]]
"""A callable type that represents the auxiliary decryption scheme."""



# support for computing discrete logs, with a cache so they're never recomputed

import asyncio
from typing import Dict, Optional

from .group import G, ElementModP, ONE_MOD_P, mult_p, int_to_p_unchecked

__dlog_cache: Dict[ElementModP, int] = {ONE_MOD_P: 0}
__dlog_max_elem = ONE_MOD_P
__dlog_max_exp = 0

__dlog_lock: Optional[asyncio.Lock] = None


def discrete_log(e: ElementModP) -> int:
    """
    Computes the discrete log (base g, mod p) of the given element,
    with internal caching of results. Should run efficiently when called
    multiple times when the exponent is at most in the single-digit millions.
    Performance will degrade if it's much larger.

    Note: *this function is thread-safe*. For the best possible performance,
    pre-compute the discrete log of a number you expect to have the biggest
    exponent you'll ever see. After that, the cache will be fully loaded,
    and every call will be nothing more than a dictionary lookup.
    """
    global __dlog_cache

    # no need for mutually exclusive access when reading from the cache
    if e in __dlog_cache:
        return __dlog_cache[e]
    else:
        return asyncio.run(__discrete_log_internal(e))


async def __discrete_log_internal(e: ElementModP) -> int:
    global __dlog_cache
    global __dlog_max_elem
    global __dlog_max_exp
    global __dlog_lock

    if __dlog_lock is None:
        # Initialize the lock on on first function call per process
        __dlog_lock = asyncio.Lock()

    async with __dlog_lock:
        g = int_to_p_unchecked(G)
        while e != __dlog_max_elem:
            __dlog_max_exp = __dlog_max_exp + 1
            __dlog_max_elem = mult_p(g, __dlog_max_elem)
            __dlog_cache[__dlog_max_elem] = __dlog_max_exp

        return __dlog_cache[__dlog_max_elem]



from typing import TypeVar, Callable

from hypothesis.strategies import composite, integers, SearchStrategy

from electionguard.group import (
    Q,
    P,
    int_to_p_unchecked,
    int_to_q_unchecked,
)

_T = TypeVar("_T")
_DrawType = Callable[[SearchStrategy[_T]], _T]


@composite
def elements_mod_q(draw: _DrawType):
    """
    Generates an arbitrary element from [0,Q).

    :param draw: Hidden argument, used by Hypothesis.
    """
    return int_to_q_unchecked(draw(integers(min_value=0, max_value=Q - 1)))


@composite
def elements_mod_q_no_zero(draw: _DrawType):
    """
    Generates an arbitrary element from [1,Q).

    :param draw: Hidden argument, used by Hypothesis.
    """
    return int_to_q_unchecked(draw(integers(min_value=1, max_value=Q - 1)))


@composite
def elements_mod_p(draw: _DrawType):
    """
    Generates an arbitrary element from [0,P).

    :param draw: Hidden argument, used by Hypothesis.
    """
    return int_to_p_unchecked(draw(integers(min_value=0, max_value=P - 1)))


@composite
def elements_mod_p_no_zero(draw: _DrawType):
    """
    Generates an arbitrary element from [1,P).

    :param draw: Hidden argument, used by Hypothesis.
    """
    return int_to_p_unchecked(draw(integers(min_value=1, max_value=P - 1)))



import os
from jsons import KEY_TRANSFORMER_SNAKECASE, loads
from random import Random, randint
from typing import cast, TypeVar, Callable, List, Tuple
import uuid

from hypothesis.strategies import (
    composite,
    booleans,
    text,
    uuids,
    SearchStrategy,
)

from electionguard.ballot import (
    PlaintextBallot,
    PlaintextBallotContest,
    PlaintextBallotSelection,
)

from electionguard.election import (
    ContestDescription,
    SelectionDescription,
    InternalElectionDescription,
)

from electionguard.encrypt import selection_from

_T = TypeVar("_T")
_DrawType = Callable[[SearchStrategy[_T]], _T]

data = os.path.realpath(os.path.join(__file__, "../../../data"))


class BallotFactory(object):
    simple_ballot_filename = "ballot_in_simple.json"
    simple_ballots_filename = "plaintext_ballots_simple.json"

    def get_random_selection_from(
        self,
        description: SelectionDescription,
        random_source: Random,
        is_placeholder=False,
    ) -> PlaintextBallotSelection:

        selected = bool(random_source.randint(0, 1))
        return selection_from(description, is_placeholder, selected)

    def get_random_contest_from(
        self,
        description: ContestDescription,
        random: Random,
        suppress_validity_check=False,
        with_trues=False,
    ) -> PlaintextBallotContest:
        """
        Get a randomly filled contest for the given description that 
        may be undervoted and may include explicitly false votes.
        Since this is only used for testing, the random number generator
        (`random`) must be provided to make this function deterministic.
        """
        if not suppress_validity_check:
            assert description.is_valid(), "the contest description must be valid"

        selections: List[PlaintextBallotSelection] = list()

        voted = 0

        for selection_description in description.ballot_selections:
            selection = self.get_random_selection_from(selection_description, random)
            # the caller may force a true value
            voted += selection.to_int()
            if voted <= 1 and selection.to_int() and with_trues:
                selections.append(selection)
                continue

            # Possibly append the true selection, indicating an undervote
            if voted <= description.number_elected and bool(random.randint(0, 1)) == 1:
                selections.append(selection)
            # Possibly append the false selections as well, indicating some choices
            # may be explicitly false
            elif bool(random.randint(0, 1)) == 1:
                selections.append(selection_from(selection_description))

        return PlaintextBallotContest(description.object_id, selections)

    def get_fake_ballot(
        self,
        election: InternalElectionDescription,
        ballot_id: str = None,
        with_trues=True,
    ) -> PlaintextBallot:
        """
        Get a single Fake Ballot object that is manually constructed with default vaules
        """

        if ballot_id is None:
            ballot_id = "some-unique-ballot-id-123"

        contests: List[PlaintextBallotContest] = []
        for contest in election.get_contests_for(election.ballot_styles[0].object_id):
            contests.append(
                self.get_random_contest_from(contest, Random(), with_trues=with_trues)
            )

        fake_ballot = PlaintextBallot(
            ballot_id, election.ballot_styles[0].object_id, contests
        )

        return fake_ballot

    def generate_fake_plaintext_ballots_for_election(
        self, election: InternalElectionDescription, number_of_ballots: int
    ) -> List[PlaintextBallot]:
        ballots: List[PlaintextBallot] = []
        for i in range(number_of_ballots):

            style_index = randint(0, len(election.ballot_styles) - 1)
            ballot_style = election.ballot_styles[style_index]
            ballot_id = f"ballot-{uuid.uuid1()}"

            contests: List[PlaintextBallotContest] = []
            for contest in election.get_contests_for(ballot_style.object_id):
                contests.append(
                    self.get_random_contest_from(contest, Random(), with_trues=True)
                )

            ballots.append(PlaintextBallot(ballot_id, ballot_style.object_id, contests))

        return ballots

    def get_simple_ballot_from_file(self) -> PlaintextBallot:
        return self._get_ballot_from_file(self.simple_ballot_filename)

    def get_simple_ballots_from_file(self) -> List[PlaintextBallot]:
        return self._get_ballots_from_file(self.simple_ballots_filename)

    def _get_ballot_from_file(self, filename: str) -> PlaintextBallot:
        with open(os.path.join(data, filename), "r") as subject:
            result = subject.read()
            target = PlaintextBallot.from_json(result)
        return target

    def _get_ballots_from_file(self, filename: str) -> List[PlaintextBallot]:
        with open(os.path.join(data, filename), "r") as subject:
            result = subject.read()
            target = cast(
                List[PlaintextBallot],
                loads(
                    result,
                    List[PlaintextBallot],
                    key_transformer=KEY_TRANSFORMER_SNAKECASE,
                ),
            )
        return target


@composite
def get_selection_well_formed(
    draw: _DrawType, uuids=uuids(), bools=booleans(), text=text()
) -> Tuple[str, PlaintextBallotSelection]:
    use_none = draw(bools)
    if use_none:
        extra_data = None
    else:
        extra_data = draw(text)
    object_id = f"selection-{draw(uuids)}"
    return (
        object_id,
        PlaintextBallotSelection(object_id, f"{draw(bools)}", draw(bools), extra_data),
    )


@composite
def get_selection_poorly_formed(
    draw: _DrawType, uuids=uuids(), bools=booleans(), text=text()
) -> Tuple[str, PlaintextBallotSelection]:
    use_none = draw(bools)
    if use_none:
        extra_data = None
    else:
        extra_data = draw(text)
    object_id = f"selection-{draw(uuids)}"
    return (
        object_id,
        PlaintextBallotSelection(object_id, f"{draw(text)}", draw(bools), extra_data),
    )
